﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Domain.DTO;
using Domain.DTO.DialogueModels;
using System.Diagnostics;

namespace Domain.Implementation
{
    public class INNk : ICcbrAlgorithm
    {
        public int K { get; set; }
        public string LocalOrGlobal { get; set; }
        public List<List<string>> AttributesWithPossibleValues { get; set; }
        public List<DataAttribute> CurrentQuery { get; set; }
        public DataClass TargetClass { get; set; }
        public bool ClassFound { get; set; }
        public List<DataCase> CaseBase { get; set; }
        public List<DataCase> TestSet { get; set; }
        public List<DataCase> RetrievalSet { get; set; }
        public DataCase CurrentCase { get; set; }
        public List<RelatedAttributesSimulated> AttributeRelations { get; set; }

        private readonly DataSet _dataSet;
        private double _progression;
        private int _similartyFunction;
        private int _attributeSelectionFunction;
        private int _inferenceMethod = 0;

        public INNk(DataSet dataSet, int k, string localOrGlobal, int simFunc = 0, int attrSelectFunc = 0, int inferenceMethod = 0)
        {
            _dataSet = dataSet;
            CaseBase = DataCaseRepository.GetAllForDataset(_dataSet.Id, true);

            TestSet = new List<DataCase>();
            if (CaseBase.Any(x => x.TrainingCase == true))
            {
                TestSet = CaseBase.Where(x => x.TrainingCase == false).ToList();
                CaseBase.RemoveAll(x => x.TrainingCase == false);
            }

            K = k;
            LocalOrGlobal = localOrGlobal;
            _similartyFunction = simFunc;
            _attributeSelectionFunction = attrSelectFunc;
            ClassFound = false;
            CurrentCase = new DataCase();

            AttributesWithPossibleValues = GetAttributesWithPossibleValues(CaseBase);
            
            _inferenceMethod = inferenceMethod;

            AttributeRelations = DataAttributeRepository.GetAttributeRelations(_dataSet.AttributeTypes.Select(x => x.Id).ToList());
        }

        public double GetProgression()
        {
            return _progression;
        }

        public double CalculateSimilarty(DataCase storedCase)
        {
            if (_similartyFunction == 1)
            {
                return SimilartyFunctionNaCoDAE(storedCase);
            }

            var sim = 0;
            foreach (var attr in CurrentQuery.Where(x => x.Value != "" && x.Value != "?"))
            {
                var storedAttr = storedCase.Attributes.FirstOrDefault(x => x.AttributeType.Position == attr.AttributeType.Position);
                if (storedAttr.Value == "?") continue;
                if (storedAttr.Value == attr.Value) sim++;
            }
            
            return Math.Round(((double)sim / CurrentQuery.Count), 5);
        }

        public double SimilartyFunctionNaCoDAE(DataCase storedCase)
        {
            var sameAnswers = 0.0;
            var differentAnswers = 0.0;
            foreach (var attr in CurrentQuery)
            {
                if (attr.Value == "") continue;
                var caseAttr = storedCase.Attributes.FirstOrDefault(x => x.AttributeType.Position == attr.AttributeType.Position);
                if (attr.AttributeType.Type == (int)AttributeEnumType.Bool || attr.AttributeType.Type == (int)AttributeEnumType.MultiSelect)
                {
                    if (caseAttr.Value == "?") continue;
                    if (attr.Value == caseAttr.Value)
                    {
                        sameAnswers++;
                    }
                    else
                    {
                        differentAnswers++;
                    }
                }
                else if (attr.AttributeType.Type == (int)AttributeEnumType.Double || attr.AttributeType.Type == (int)AttributeEnumType.Integer)
                {
                    var minValue = attr.AttributeType.MinRange;
                    var maxValue = attr.AttributeType.MaxRange;
                    var maxDifference = (double)maxValue - minValue;
                    var caseValue = 0.0;
                    var queryValue = 0.0;
                    if (caseAttr.Value == "?") continue;
                    double.TryParse(attr.Value.Replace(',', '.'), out queryValue);
                    double.TryParse(caseAttr.Value.Replace(',', '.'), out caseValue);
                    var querySimilarity = 1 - ((double)Math.Abs((queryValue - caseValue)) / maxDifference);
                    sameAnswers += querySimilarity;
                }
            }

            return ((sameAnswers - differentAnswers) / storedCase.Attributes.Count);
        }


        public List<ScoreObject> FindNewRetrievalSet(int caseId)
        {
            RetrievalSet = new List<DataCase>();
            var similiartyMap = CaseBase.ToDictionary(storedCase => storedCase, CalculateSimilarty);
            similiartyMap = similiartyMap.OrderBy(x => x.Value).ToDictionary(x => x.Key, x => x.Value);
            var current = similiartyMap.FirstOrDefault(x => x.Key.Id == caseId);
            if (current.Key != null)
            {
                similiartyMap.Remove(current.Key);
            
            }

            var returnList = new List<ScoreObject>();
            foreach (var caseSim in similiartyMap)
            {
                if (caseSim.Key.Id != caseId)
                {
                    var count = similiartyMap.Count(otherCase => otherCase.Value > caseSim.Value);
                    if (count <= K)
                    {
                        RetrievalSet.Add(caseSim.Key);
                        returnList.Add(new ScoreObject { Case = caseSim.Key, Score = caseSim.Value });
                    }
                }
            }

            return returnList;
        }

        public List<List<string>> GetAttributesWithPossibleValues(List<DataCase> dataSet)
        {
            var returnList = new List<List<string>>();
            for (var i = 0; i < _dataSet.AttributeTypes.Count; i++)
            {
                returnList.Add(new List<string>());
                foreach (var storedCase in dataSet)
                {
                    if (storedCase.Id == CurrentCase.Id) continue;
                    var attribute = storedCase.Attributes.FirstOrDefault(x => x.AttributeType.Id == _dataSet.AttributeTypes.ElementAt(i).Id);
                    var attributeValue = attribute.Value;
                    if (attributeValue == "?") continue;
                    if (!returnList.ElementAt(i).Contains(attributeValue))
                    {
                        returnList.ElementAt(i).Add(attributeValue);
                    }
                }
            }
            return returnList;
        }

        public bool OnlyMissingValuesLeft(List<DataCase> cases)
        {
            var attrIndex = 0;
            foreach (var attr in CurrentQuery)
            {
                if (attr.Value == "")
                {
                    if (cases.Any(ca => ca.Attributes.ElementAt(attrIndex).Value != "?"))
                    {
                        return false;
                    }
                }
                attrIndex++;
            }
            return true;
        }

        public DataClass SelectNewTargetClass()
        {
            var classGroups = RetrievalSet.GroupBy(x => x.DataClass).OrderByDescending(x => x.Count());

            if (classGroups.Count() == 1)
            {
                ClassFound = true;
                return classGroups.First().First().DataClass;
            }
            if (CurrentQuery.Count(x => x.Value != "") == CurrentQuery.Count())
            {
                ClassFound = true;
                if (K != 1)
                {
                    var temp = K;
                    K = 1;
                    FindNewRetrievalSet(CurrentCase.Id);
                    classGroups = RetrievalSet.GroupBy(x => x.DataClass).OrderByDescending(x => x.Count());
                    K = temp;
                }
            }

            var highestCount = classGroups.First().Count();
            var listOfMostSupported = new List<DataClass>();

            foreach (var group in classGroups)
            {
                if (group.Count() == highestCount)
                {
                    listOfMostSupported.Add(group.First().DataClass);
                }
            }

            if (listOfMostSupported.Count > 1)
            {
                var caseBaseClassGroups = CaseBase.GroupBy(x => x.DataClass).OrderByDescending(x => x.Count());
                foreach (var group in caseBaseClassGroups)
                {
                    if (listOfMostSupported.Contains(group.First().DataClass))
                    {
                        return group.First().DataClass;
                    }
                }
            }

            return listOfMostSupported.First();
        }

        public List<AttributeValuePowerTripplet> CalculateDiscriminatingPowerForAttribute(int attributeIndex)
        {
            var returnList = new List<AttributeValuePowerTripplet>();
            var attribute = _dataSet.AttributeTypes.ElementAt(attributeIndex);
            var attrValues = AttributesWithPossibleValues.ElementAt(attributeIndex);
            foreach (var value in attrValues)
            {
                var valueGivenClass = 0.0;
                var classCounter = 0.0;
                var notClassCounter = 0.0;
                var valueGivenNotClass = 0.0;

                var listToUse = RetrievalSet;
                if (LocalOrGlobal == "global")
                {
                    listToUse = CaseBase;
                }

                foreach (var ca in listToUse)
                {
                    if (ca.Id == CurrentCase.Id) continue;
                    var caseAttrValue = ca.Attributes.ElementAt(attributeIndex).Value;

                    if (ca.DataClass == TargetClass)
                    {
                        classCounter++;
                        if (caseAttrValue == value)
                        {
                            valueGivenClass++;
                        }
                    }
                    else
                    {
                        notClassCounter++;
                        if (caseAttrValue == value)
                        {
                            valueGivenNotClass++;
                        }
                    }
                }
                var d = ((valueGivenClass / classCounter) - (valueGivenNotClass / notClassCounter)) / (attrValues.Count);

                returnList.Add(new AttributeValuePowerTripplet { Attribute = attribute, AttributeIndex = attributeIndex, Value = value, DiscriminatingPower = d });

            }

            return returnList;
        }

        public int FindNextAttributeIndex()
        {
            var attrIndex = 0;
            var attrWithDiscPower = new List<AttributeValuePowerTripplet>();
            foreach (var attr in CurrentQuery)
            {
                if (attr.Value == "")
                {
                    attrWithDiscPower.AddRange(CalculateDiscriminatingPowerForAttribute(attrIndex));
                }
                attrIndex++;
            }

            attrWithDiscPower = attrWithDiscPower.OrderByDescending(x => x.DiscriminatingPower).ToList();

            var listToUse = RetrievalSet;
            if (LocalOrGlobal == "global")
            {
                listToUse = CaseBase;
            }

            foreach (var attr in attrWithDiscPower)
            {
                if (listToUse.Any(x => (x.DataClass == TargetClass) && (x.Attributes.ElementAt(attr.AttributeIndex).Value == attr.Value)))
                {
                    return attr.AttributeIndex;
                }
            }

            return 0;

        }

        public void SaveAnswer(DataAttribute question)
        {
            CurrentQuery.First(d => d.AttributeType.Id == question.AttributeType.Id).Value = question.Value;

        }

        public DialogueStateModel InitializeDialog(List<DataAttribute> answeredQuestions)
        {

            var returnModel = new DialogueStateModel();
            ClassFound = false;

            RetrievalSet = CaseBase.Where(x => x.Id != CurrentCase.Id).ToList();

            CurrentQuery = new List<DataAttribute>();
            foreach (var at in _dataSet.AttributeTypes)
            {
                CurrentQuery.Add(new DataAttribute { AttributeType = at, Value = "" });
                foreach (var ans in answeredQuestions)
                {
                    var relatedAttribute = ans.AttributeType.RelatedAttributes.FirstOrDefault(x => x.Id == at.Id);
                    if (relatedAttribute != null)
                    {
                        CurrentQuery.Last().Value = ans.Value;
                        break;
                    }
                }
            }
            if (CurrentQuery.Any(x => x.Value != ""))
            {
                returnModel.ScoreObjects = FindNewRetrievalSet(CurrentCase.Id);
            }
            TargetClass = SelectNewTargetClass();
            if (OnlyMissingValuesLeft(RetrievalSet.Where(x => x.DataClass == TargetClass).ToList()))
            {
                ClassFound = true;
                if (K != 1)
                {
                    var temp = K;
                    K = 1;
                    FindNewRetrievalSet(CurrentCase.Id);
                    TargetClass = SelectNewTargetClass();
                    K = temp;
                }
                returnModel.Question = null;
                returnModel.DataClass = TargetClass;
                return returnModel;
            }
            var nextQuestion = new AttributeType();
            if (_attributeSelectionFunction == 1)
            {
                nextQuestion = SelectionStrategies.GetBestAttributeByEntropy(RetrievalSet, CurrentQuery.Where(x => x.Value == "").Select(x => x.AttributeType).ToList(), _dataSet);
            }
            else if (_attributeSelectionFunction == 2)
            {
                nextQuestion = SelectionStrategies.GetBestAttributeHeterogenous(_dataSet, RetrievalSet, CurrentQuery.Where(x => x.Value != "").Select(y => y.AttributeType).ToList());
            }
            else
            {
                var attrIndex = FindNextAttributeIndex();
                nextQuestion = _dataSet.AttributeTypes.ElementAt(attrIndex);
            }
            returnModel.CurrentQuery = CurrentQuery;
            returnModel.Question = nextQuestion;
            returnModel.DataClass = TargetClass;

            return returnModel;
        }

        public DialogueStateModel GetNextQuestion()
        {
            var caseSimilartyScores = FindNewRetrievalSet(CurrentCase.Id);

            TargetClass = SelectNewTargetClass();

            if (!ClassFound)
            {
                if (OnlyMissingValuesLeft(RetrievalSet.Where(x => x.DataClass == TargetClass).ToList()))
                {
                    ClassFound = true;
                    if (K != 1)
                    {
                        var temp = K;
                        K = 1;
                        FindNewRetrievalSet(CurrentCase.Id);
                        TargetClass = SelectNewTargetClass();
                        K = temp;
                    }
                }
            }

            if (ClassFound)
            {
                return new DialogueStateModel
                {
                    DataCasesWithSameAttributes = new List<DataCase>(),
                    NumberOfAskedQuestions = CurrentQuery.Count(x => x.Value != ""),
                    Question = null,
                    DataClass = TargetClass,
                    ScoreObjects = caseSimilartyScores,
                    CurrentQuery = CurrentQuery
                };
            }
            var remainingAttributes = CurrentQuery.Where(x => x.Value == "").Select(x => x.AttributeType).ToList();
            var question = new AttributeType();

            var listToUse = RetrievalSet;
            if (LocalOrGlobal == "global")
            {
                listToUse = CaseBase;
            }

            if (_attributeSelectionFunction == 1)
            {
                question = SelectionStrategies.GetBestAttributeByEntropy(listToUse, remainingAttributes, _dataSet);
            }
            else if (_attributeSelectionFunction == 2)
            {
                question = SelectionStrategies.GetBestAttributeHeterogenous(_dataSet, listToUse, CurrentQuery.Where(x => x.Value != "").Select(y => y.AttributeType).ToList());
            }
            else
            {
                question = _dataSet.AttributeTypes.ElementAt(FindNextAttributeIndex());
            }

            return new DialogueStateModel
            {
                DataCasesWithSameAttributes = new List<DataCase>(),
                NumberOfAskedQuestions = CurrentQuery.Count(x => x.Value != ""),
                Question = question,
                DataClass = TargetClass,
                ScoreObjects = caseSimilartyScores,
                CurrentQuery = CurrentQuery
            };

        }

        public LeaveOneOutResultModel LeaveOneOutTest()
        {
            var correct = 0;
            var numberOfCases = 0;
            var numOfQuestionsAsked = 0.0;
            var longestDialogue = 0;
            var shortestDialogue = _dataSet.AttributeTypes.Count();
            var testSet = TestSet.Any() ? TestSet : CaseBase;

            //for (var i = 1; i < 6; i++)
            //{
            //    K = i;
            //    _progression = 0.0;
            //    correct = 0;
            //    numberOfCases = 0;
            //    numOfQuestionsAsked = 0.0;
            //    longestDialogue = 0;
            //    shortestDialogue = _dataSet.AttributeTypes.Count();

                foreach (var cs in testSet)
                {
                    CurrentCase = cs;
                    var qaModel = InitializeDialog(new List<DataAttribute>());
                    var q = 0;
                    while (!ClassFound)
                    {
                        var index = _dataSet.AttributeTypes.IndexOf(qaModel.Question);
                        var answer = new DataAttribute
                        {
                            AttributeType = qaModel.Question,
                            Value = CurrentCase.Attributes.ElementAt(index).Value
                        };
                        SaveAnswer(answer);
                        q++;
                        if (_inferenceMethod == 1)
                        {
                            CurrentQuery = InferenceModule.InferAnswers(_dataSet.Id,
                                CurrentCase.Attributes, _dataSet.AttributeTypes);
                        }
                        else if (_inferenceMethod == 2)
                        {
                            var inferedAnswers = InferenceModule.InferAnswersStatisticly(AttributeRelations,
                                CurrentQuery.Where(x => x.Value != "").ToList(), _dataSet.AttributeTypes, answer);

                            foreach (var ans in inferedAnswers)
                            {
                                if (!CurrentQuery.Any(x => x.Id == ans.AttributeType.Id))
                                {
                                    CurrentQuery.Add(ans);
                                }
                                else
                                {
                                    CurrentQuery.FirstOrDefault(
                                        x => x.AttributeType.Id == ans.AttributeType.Id).Value = ans.Value;
                                }
                            }
                        }


                        qaModel = GetNextQuestion();
                    }


                    if (CurrentCase.DataClass == TargetClass)
                    {
                        correct++;
                    }
                    

                    var dialogueLenght = q;
                    if (dialogueLenght > longestDialogue) longestDialogue = dialogueLenght;
                    if (dialogueLenght < shortestDialogue) shortestDialogue = dialogueLenght;
                    numOfQuestionsAsked += dialogueLenght;
                    numberOfCases++;
                    _progression = Math.Round(((double)numberOfCases / testSet.Count) * 100, 2);
                }
            //For writing results to a file. Commented out because we dont want the server to write to a file. Slows down the system for the user
            /*
                var basePath = AppDomain.CurrentDomain.BaseDirectory;
                var path = Path.Combine(basePath, "Content/Logs/INNk.txt");
                var appEndString = "";
                if (File.Exists(path))
                {
                    using (StreamReader sr = File.OpenText(path))
                    {
                        appEndString = sr.ReadToEnd();
                    }
                    File.Delete(path);
                    // Create a file to write to. 

                }
                using (StreamWriter sw = File.CreateText(path))
                {
                    sw.WriteLine(appEndString);
                    sw.WriteLine("iNN(k)-" + LocalOrGlobal.Substring(0, 1).ToUpper());
                    sw.WriteLine("DataSet: " + _dataSet.Name);
                    sw.WriteLine("Number of cases tested: " + testSet.Count);
                    sw.WriteLine("Date: " + DateTime.Now);
                    sw.WriteLine("K: " + K);
                    sw.WriteLine("Selection strategy: " + (_attributeSelectionFunction == 1 ? "Entropy" : "Discriminating power"));
                    sw.WriteLine("Similarity measure: " + (_similartyFunction == 1 ? "NaCoDAE" : "iNN(k)"));
                    sw.WriteLine("Correct : " + (double)correct / numberOfCases);
                    sw.WriteLine("Effectiveness: " + numOfQuestionsAsked / (_dataSet.AttributeTypes.Count * numberOfCases));
                    sw.WriteLine("Total number of questions: " + numOfQuestionsAsked);
                    sw.WriteLine("Shortest dialogue: " + shortestDialogue);
                    sw.WriteLine("Longest dialogue: " + longestDialogue);
                    sw.WriteLine("");
                }
             */
            
            _progression = 100.0;
            
            return new LeaveOneOutResultModel
            {
                CaseBaseCount = testSet.Count,
                K = K,
                LocalOrGlobal = LocalOrGlobal,
                AttributeSelectionFunction = (_attributeSelectionFunction == 1 ? "Entropy" : "Discriminating power"),
                SimilartyFunction = (_similartyFunction == 1 ? "From NaCoDAE" : "From iNN(k)"),
                PercentCorrect = (double)correct / numberOfCases,
                Effectiveness = numOfQuestionsAsked / (_dataSet.AttributeTypes.Count * numberOfCases),
                NumberOfMatchingCases = correct,
                Shortest = shortestDialogue,
                Longest = longestDialogue,
                NumberOfQuestionsAsked = (numOfQuestionsAsked / (_dataSet.AttributeTypes.Count * numberOfCases)) * _dataSet.AttributeTypes.Count
            };
        }

    }
    public class AttributeValuePowerTripplet
    {
        public AttributeType Attribute { get; set; }
        public int AttributeIndex { get; set; }
        public string Value { get; set; }
        public double DiscriminatingPower { get; set; }
    }
}
