﻿using System.IO;
using Domain.DTO;
using Domain.DTO.DialogueModels;

namespace Domain.Implementation
{
    using System;
    using System.Collections.Generic;
    using System.Linq;


    namespace Master.Implementation
    {
        public class NaCoDAE : ICcbrAlgorithm
        {
            public List<DataCase> CaseBase { get; set; }
            public List<DataCase> TestCaseBase { get; set; }
            public int MaxSymptoms { get; set; }
            public DataCase CurrentCase { get; set; }
            public List<DataCase> RetrievalSet { get; set; }
            public List<ScoreObject> Scores { get; set; }
            public List<AttributeType> CurrentQuery { get; set; }
            public List<RelatedAttributesSimulated> AttributeRelations { get; set; }

            private DataSet _dataSet;
            private DataClass _currentBestClass;
            private List<ScoreObject> _currentBestScores;
            private double progression = 0;
            private double Alpa = 0.5;
            private int _retrievalSetSize = 5;
            private double _thresholdValue = 0.4;
            private int _thresholdFunction = 0;
            private int _similartyFunction;
            private int _attributeSelectionFunction;
            private int _inferenceMethod = 0;

            public NaCoDAE(DataSet dataSet, int retrievalSetSize = 5, double thresholdValue = 0.4, int  thresholdFunction = 0, int simFunc = 1, int attrSelectFunc = 2, int inferenceMethod = 0)
            {
                _dataSet = dataSet;
                CaseBase = DataCaseRepository.GetAllForDataset(dataSet.Id, false).OrderBy(x => Guid.NewGuid()).ToList();

                TestCaseBase = new List<DataCase>();
                if (CaseBase.Any(x => x.TrainingCase == true))
                {
                    TestCaseBase = CaseBase.Where(x => x.TrainingCase == false).ToList();
                    CaseBase.RemoveAll(x => x.TrainingCase == false);
                }

                MaxSymptoms = _dataSet.AttributeTypes.Count;
                _retrievalSetSize = retrievalSetSize;
                _similartyFunction = simFunc;
                _attributeSelectionFunction = attrSelectFunc;
                _thresholdValue = thresholdValue;
                _thresholdFunction = thresholdFunction;
                _inferenceMethod = inferenceMethod;

                AttributeRelations = DataAttributeRepository.GetAttributeRelations(_dataSet.AttributeTypes.Select(x => x.Id).ToList());
            }

            public double GetProgression()
            {
                return progression;
            }

            public void QuerySimilarity(bool leaveCurrentCaseOut = false)
            {
                var scores = new List<ScoreObject>();

                foreach (var patientCase in CaseBase)
                {

                    if (_similartyFunction == 2)
                    {
                        var sim = CurrentCase.Attributes.Where(t=> t.Value != "" && patientCase.Attributes.FirstOrDefault(y => t.AttributeType.Id == t.AttributeType.Id).Value == t.Value).Count();
                        scores.Add(new ScoreObject { Case = patientCase, Score = (double)sim / patientCase.Attributes.Count() });
                        continue;
                    }

                    var sameAnswers = 0.0;
                    var differentAnswers = 0.0;
                    
                    if (patientCase.Id == CurrentCase.Id && leaveCurrentCaseOut) continue;
                    
                    foreach (var attribute in CurrentQuery)
                    {
                        var patientAttribute = patientCase.Attributes.FirstOrDefault(x => x.AttributeType.Id == attribute.Id);
                        if (patientAttribute == null || patientAttribute.Value == "?") continue;

                        var currentQueryBaseAttribute = CurrentCase.Attributes.FirstOrDefault(x => x.AttributeType.Id == attribute.Id);
                        if (currentQueryBaseAttribute == null) continue;
                        
                        if (attribute.Type == (int)AttributeEnumType.Bool || attribute.Type == (int)AttributeEnumType.MultiSelect)
                        {
                            if (currentQueryBaseAttribute.Value == patientAttribute.Value) sameAnswers++;
                            else differentAnswers++;
                        }
                        else if (attribute.Type == (int)AttributeEnumType.Double || attribute.Type == (int)AttributeEnumType.Integer)
                        {
                            if (_similartyFunction == 0)
                            {
                                if (currentQueryBaseAttribute.Value == patientAttribute.Value) sameAnswers++;
                                else differentAnswers++;
                            }
                            else
                            {
                                var minValue = attribute.MinRange;
                                var maxValue = attribute.MaxRange;
                                var maxDifference = (double)maxValue - minValue;

                                var patientValue = 0.0;
                                var queryValue = 0.0;
                                double.TryParse(currentQueryBaseAttribute.Value.Replace(',', '.'), out queryValue);
                                double.TryParse(patientAttribute.Value.Replace(',', '.'), out patientValue);

                                var querySimilarity = 1 - (Math.Abs((queryValue - patientValue)) / maxDifference);
                                sameAnswers += querySimilarity;
                            }
                        }
                    }

                    var queryScore = (sameAnswers - differentAnswers) / patientCase.Attributes.Count();
                 
                    var unbiasedScore = GetUnbiasedScore(queryScore, Math.Max(CurrentQuery.Count(), patientCase.Attributes.Count(x => x.Value != "?")), Alpa);

                    scores.Add(new ScoreObject { Case = patientCase, Score = queryScore });
                }

                Scores = scores;
            }

            public int CheckIfEqual(int attributeType, string value, string value2)
            {
                if (value == "?" || value2 == "?") return 0;
                if (attributeType == (int)AttributeEnumType.Bool)
                {
                    return value == value2 ? 1 : -1;
                }
                
                if (attributeType == (int)AttributeEnumType.Double || attributeType == (int)AttributeEnumType.Integer)
                {
                    double patientValue;
                    double queryValue;

                    double.TryParse(value.Replace('.', ','), out queryValue);
                    double.TryParse(value2.Replace('.', ','), out patientValue);

                    if (patientValue == queryValue) return 1;

                    if (patientValue == 0.0 || queryValue == 0.0) return 0;

                    return -1;
                }

                return 0;
            }

            public double GetUnbiasedScore(double queryScore, int maxAttr, double alpha)
            {
                return (alpha * Math.Max(0, queryScore) + ((1 - alpha) * maxAttr)) / (alpha + ((1 - alpha) * maxAttr));
            }

            public AttributeType GetNextQuestionIndex(List<DataCase> cases, List<AttributeType> query)
            {
                if (query.Count > _dataSet.AttributeTypes.Count) return null;

                var remainingAttributes = _dataSet.AttributeTypes.Where(x => !query.Contains(x)).ToList();
                if (_attributeSelectionFunction == 1) {
                    return SelectionStrategies.GetBestAttributeByEntropy(cases, remainingAttributes, _dataSet);
                }
                else if (_attributeSelectionFunction == 2)
                {
                    return SelectionStrategies.GetBestAttributeHeterogenous(_dataSet, RetrievalSet, CurrentQuery);
                }
                return SelectionStrategies.GetBestAttributeByMagic(_dataSet, _currentBestClass, _currentBestScores, RetrievalSet, cases, query);
       
            }


            public DialogueStateModel InitializeDialog(List<DataAttribute> answeredQuestions)
            {
                RetrievalSet = CaseBase;
                CurrentCase = new DataCase { Attributes = new List<DataAttribute>() };
                CurrentQuery = new List<AttributeType>();

                foreach (var at in _dataSet.AttributeTypes)
                {
                    
                    foreach (var ans in answeredQuestions)
                    {
                        var relatedAttribute = ans.AttributeType.RelatedAttributes.FirstOrDefault(x => x.Id == at.Id);
                        if (relatedAttribute != null)
                        {
                            CurrentQuery.Add(at);
                            CurrentCase.Attributes.Add(ans);
                            break;
                        }
                    }
                }


                var nextQuestion = GetNextQuestionIndex(RetrievalSet, CurrentQuery);
                var returnModel = new DialogueStateModel { Question = nextQuestion, DataClass = null };

                returnModel.CurrentQuery = CurrentCase.Attributes;

                return returnModel;
            }

            public void SaveAnswer(DataAttribute question)
            {
                CurrentQuery.Add(question.AttributeType);

                //if (((double)CurrentQuery.Count / _dataSet.AttributeTypes.Count) < 0.3) _retrievalSetSize = CaseBase.Count;
                if (CurrentCase.Attributes == null) CurrentCase.Attributes = new List<DataAttribute>();

                CurrentCase.Attributes.Add(question);
                QuerySimilarity();
                var newRetrievelSet = Scores.OrderByDescending(x => x.Score).Take(_retrievalSetSize).Select(x => new { x.Case, x.Score }).ToList();

                if (!newRetrievelSet.Any()) return;
                _currentBestClass = newRetrievelSet.FirstOrDefault().Case.DataClass;
                _currentBestScores = Scores.OrderByDescending(x => x.Score).Take(_retrievalSetSize).ToList();
                RetrievalSet = newRetrievelSet.Select(x => x.Case).Take(_retrievalSetSize).ToList();
            }


            public DialogueStateModel GetNextQuestion()
            {
                var nextQuestion = GetNextQuestionIndex(RetrievalSet, CurrentQuery);
                var returnModel = new DialogueStateModel();

                returnModel.DataCasesWithSameAttributes = new List<DataCase>();
                var numberOfAskedQuestions = CurrentQuery.Count;
                returnModel.NumberOfAskedQuestions = numberOfAskedQuestions;
                foreach (var dataCase in CaseBase)
                {
                    var numberOfRightCases = 0;
                    foreach (var attributeType in CurrentQuery)
                    {
                        var selectedAttribute =
                            dataCase.Attributes.FirstOrDefault(x => x.AttributeType.Id == attributeType.Id);
                        var savedAnswer =
                            CurrentCase.Attributes.FirstOrDefault(x => x.AttributeType.Id == attributeType.Id);

                        if (savedAnswer != null && selectedAttribute != null)
                        {
                            var result = CheckIfEqual(attributeType.Type, selectedAttribute.Value, savedAnswer.Value);
                            if (result == 1) numberOfRightCases++;
                        }
                    }
                    if (numberOfRightCases == numberOfAskedQuestions) returnModel.DataCasesWithSameAttributes.Add(dataCase);
                }

                returnModel.Question = nextQuestion;
                returnModel.CurrentQuery = CurrentCase.Attributes;
                returnModel.DataClass = _currentBestClass;
                returnModel.ScoreObjects = _currentBestScores;
                return returnModel;
            }


            public LeaveOneOutResultModel LeaveOneOutTest()
            {
                var returnModel = new LeaveOneOutResultModel();
                var testCases = TestCaseBase.Any() ? TestCaseBase : CaseBase;

                var correct = 0;
                var numOfQuestionsAsked = 0.0;
                var numberOfCasesRunned = 0;
                var longestDialogue = 0;
                var shortestDialogue = _dataSet.AttributeTypes.Count();

                for (var i = 0; i < 3; i++)
                {
                    CaseBase = CaseBase.OrderBy(x => Guid.NewGuid()).ToList();

                    foreach (var patientCase in testCases)
                    {
                        progression = Math.Round((((double)numberOfCasesRunned / testCases.Count) * 100) / 3, 2);
                        CurrentCase = new DataCase { Attributes = new List<DataAttribute>(), DataClass = patientCase.DataClass, Id = patientCase.Id };
                        RetrievalSet = CaseBase;
                        CurrentQuery = new List<AttributeType>();
                        var returnTopScore = true;
                        var q = 0;

                        while (true)
                        {
                            var nextQuestion = GetNextQuestionIndex(RetrievalSet, CurrentQuery);
                            if (nextQuestion == null) break;

                            var answer = patientCase.Attributes.FirstOrDefault(x => x.AttributeType.Id == nextQuestion.Id);
                            CurrentQuery.Add(nextQuestion);
                            CurrentCase.Attributes.Add(answer);

                            if (_inferenceMethod == 1)
                            {
                                CurrentCase.Attributes = InferenceModule.InferAnswers(_dataSet.Id,
                                    CurrentCase.Attributes, _dataSet.AttributeTypes);
                                foreach (var attr in CurrentCase.Attributes)
                                {
                                    if (!CurrentQuery.Any(x => x.Id == attr.AttributeType.Id))
                                    {
                                        CurrentQuery.Add(attr.AttributeType);
                                    }
                                }
                            }
                            else if (_inferenceMethod == 2)
                            {
                                var inferedAnswers = InferenceModule.InferAnswersStatisticly(AttributeRelations,
                                    CurrentCase.Attributes, _dataSet.AttributeTypes, answer);

                                foreach (var ans in inferedAnswers)
                                {
                                    if (!CurrentQuery.Any(x => x.Id == ans.AttributeType.Id))
                                    {
                                        CurrentQuery.Add(ans.AttributeType);
                                        CurrentCase.Attributes.Add(ans);
                                    }
                                    else
                                    {
                                        CurrentCase.Attributes.FirstOrDefault(
                                            x => x.AttributeType.Id == ans.AttributeType.Id).Value = ans.Value;
                                    }
                                }
                            }

                            QuerySimilarity(true);
                            numOfQuestionsAsked++;
                            q++;

                            var ordered = Scores.OrderByDescending(x => x.Score).Take(_retrievalSetSize).ToList();

                            RetrievalSet = ordered.Select(x => x.Case).ToList();

                            //Determine if dialogue should be terminated
                            if (_thresholdFunction == 0)
                            {
                                if (ScoreThresholdExceeded(_thresholdValue, ordered.FirstOrDefault().Score)) break;
                            }
                            else if (_thresholdFunction == 1)
                            {
                                if (ClassDistributionThresholdExceeded(RetrievalSet, _thresholdValue))
                                {
                                    returnTopScore = false;
                                    break;
                                }
                            }
                            else
                            {
                                if (AverageThresholdExceeded(_thresholdValue, ordered.Select(x => x.Score).ToList()))
                                    break;
                            }
                        }

                        var topScores = Scores.OrderByDescending(x => x.Score).Take(_retrievalSetSize).ToList();
                        var dataClassId = GetClassification(topScores, returnTopScore);
                        if (dataClassId == CurrentCase.DataClass.Id) correct++;

                        numberOfCasesRunned++;
                        var dialogueLenght = q;
                        if (dialogueLenght > longestDialogue) longestDialogue = dialogueLenght;
                        if (dialogueLenght < shortestDialogue) shortestDialogue = dialogueLenght;
                    }

                    var efficiency = numOfQuestionsAsked / (numberOfCasesRunned * _dataSet.AttributeTypes.Count);
                    var correctPercentage = Math.Round(((double)correct / (double)numberOfCasesRunned), 2);

                    var thresholdFunction = (ThresholdFunctions[])Enum.GetValues(typeof(ThresholdFunctions));
                    var attributeSelectionFunctions = (AttributeSelectionFunctions[])Enum.GetValues(typeof(AttributeSelectionFunctions));
                    var similartyFunctions = (SimilartyFunctions[])Enum.GetValues(typeof(SimilartyFunctions));
                    var inferenceMethod = (InferenceMethods[])Enum.GetValues(typeof(InferenceMethods));

                    //For writing results to a file. Commented out because we dont want the server to write to a file. Slows down the system for the user
                    /*
                    if (i == 2)
                    {
                        WriteResultsToFile(_dataSet.Name, _retrievalSetSize, thresholdFunction[_thresholdFunction].ToString(),
                            attributeSelectionFunctions[_attributeSelectionFunction].ToString(), similartyFunctions[_similartyFunction].ToString(),
                            inferenceMethod[_inferenceMethod].ToString(), _thresholdValue, efficiency, correctPercentage,
                            longestDialogue, shortestDialogue);
                    }
                      */      
                    returnModel = new LeaveOneOutResultModel
                    {
                        NumberOfMatchingCases = correct,
                        CaseBaseCount = testCases.Count,
                        Effectiveness = efficiency,
                        NumberOfQuestionsAsked = Math.Round(numOfQuestionsAsked / numberOfCasesRunned, 2),
                        PercentCorrect = Math.Round(((double)correct / (double)numberOfCasesRunned), 2),
                        RetrievalSetSize = _retrievalSetSize,
                        ThresholdFunction = thresholdFunction[_thresholdFunction].ToString(),
                        AttributeSelectionFunction = attributeSelectionFunctions[_attributeSelectionFunction].ToString(),
                        SimilartyFunction = similartyFunctions[_similartyFunction].ToString(),
                        ThresholdValue = _thresholdValue,
                        NumberOfQuestionsTotal = _dataSet.AttributeTypes.Count,
                        Shortest = shortestDialogue,
                        Longest = longestDialogue
                    };
                }
                    
                progression = 100;
                return returnModel;
            }

            public bool ClassDistributionThresholdExceeded(List<DataCase> retrievalSet, double threshold)
            {
                var maxOccurences = 0;
                foreach (var dataClass in _dataSet.DataClasses)
                {
                    var counter = retrievalSet.Count(dataCase => dataCase.DataClass.Id == dataClass.Id);
                    maxOccurences = Math.Max(maxOccurences, counter);
                }

                return (double) maxOccurences/retrievalSet.Count > threshold;
            }

            public bool ScoreThresholdExceeded(double threshold, double score)
            {
                return score > threshold;
            }

            public bool AverageThresholdExceeded(double threshold, List<double> retrievalSetScores)
            {
                var sum = retrievalSetScores.Sum(x => (double?)x) ?? 0;
                //var questionRounder = (1 - (questionCounter / numberOfQuestionsTotal));
                var averageScore = (sum / retrievalSetScores.Count);
                
                if (averageScore >= threshold)
                {
                    return true;
                }

                return false;
            }

            public int GetClassification(List<ScoreObject> topScores, bool returnTopScore)
            {
                var dataClassId = RetrievalSet.Select(x => x.DataClass.Id).FirstOrDefault();
                if (topScores.Count(x => x.Score == topScores.FirstOrDefault().Score) <= 1 && returnTopScore) return dataClassId;

                var tempBestScore = 0;
                
                foreach (var scoreObject in topScores.Select(x => x.Case.DataClass.Id).Distinct())
                {
                    var count = topScores.Count(x => x.Case.DataClass.Id == scoreObject);
                    if (count > tempBestScore)
                    {
                        dataClassId = scoreObject;
                        tempBestScore = count;
                    }
                }

                return dataClassId;
            }


            public void WriteResultsToFile(string datasetName, int retrievalSetSize, string terminationMethod, string selectionFunction, string simFunction, string inferenceMethod,
                double threshold, double efficiency, double correct, int longest, int shortest)
            {
                var basePath = AppDomain.CurrentDomain.BaseDirectory;
                string path = Path.Combine(basePath, "Content/Logs/Nacodae.txt");

                var appEndString = "";
                if (File.Exists(path))
                {
                    using (StreamReader sr = File.OpenText(path))
                    {
                        appEndString = sr.ReadToEnd();
                    }
                    File.Delete(path);
                }

                using (StreamWriter sw = File.CreateText(path))
                {
                    sw.WriteLine(appEndString);
                    sw.WriteLine("DataSet: " + datasetName);
                    sw.WriteLine("Date: " + DateTime.Now);
                    sw.WriteLine("Retrievel set size: " + retrievalSetSize);
                    sw.WriteLine("Termination criteria: " + terminationMethod);
                    sw.WriteLine("Attribute selection function: " + selectionFunction);
                    sw.WriteLine("Similarty function: " + simFunction);
                    sw.WriteLine("Inference method: " + inferenceMethod);
                    sw.WriteLine("Threshold value: " + threshold);
                    sw.WriteLine("PercentCorrect: " + correct);
                    sw.WriteLine("Efficiency: " + efficiency);
                    sw.WriteLine("Shortest dialogue: " + shortest);
                    sw.WriteLine("Longest dialogue: " + longest);
                    sw.WriteLine("");
                }
            }
			public List<RelatedAttributesSimulated> CheckForRelatedAttribute(int newQuestionId, string value)
            {
                using (var context = new DomainDbContext())
                {
                    var result =
                        context.RelatedAttributesSimulated.Where(
                            x => x.AttributeId == newQuestionId && x.Value == value);
                    return result.ToList();
                }
            }
            public List<Tuple<AttributeType, AttributeType, int, int, string>> RunAttributeCombiner()
            {
                var testCases = TestCaseBase.Any() ? TestCaseBase : CaseBase;

                var currentCase = new DataCase();
                var totalCases = testCases.Count;
                var tupple = new List<Tuple<AttributeType, AttributeType, int, int, string>>();
                using (var context = new DomainDbContext())
                {
                    foreach (var attribute in _dataSet.AttributeTypes)
                    {
                        var uniqueValues = context.DataAttributes.Where(x => x.DataSet.Id == _dataSet.Id && x.AttributeType.Id == attribute.Id).Select(x => x.Value).Distinct().ToList();

                        foreach (var uniqueValue in uniqueValues)
                        {
                            foreach (var attributeType in _dataSet.AttributeTypes)
                            {
                                if (attributeType.Id == attribute.Id) continue;

                                var uniqueValuesInnerList = context.DataAttributes.Where(x => x.DataSet.Id == _dataSet.Id && x.AttributeType.Id == attributeType.Id).Select(x => x.Value).Distinct().ToList();

                                var cases =
                                    testCases.Where(
                                        x =>
                                            x.Attributes.FirstOrDefault(t => t.AttributeType.Id == attribute.Id).Value ==
                                            uniqueValue).ToList();
                                var totalCount = cases.Count;
                                if (totalCount <= 1) continue;
                                foreach (var uniqueValuesInner in uniqueValuesInnerList)
                                {
                                    var testCount = 0;

                                    foreach (var dataCase in cases)
                                    {
                                        if (
                                            dataCase.Attributes.FirstOrDefault(
                                                x => x.AttributeType.Id == attributeType.Id).Value == uniqueValuesInner)
                                            testCount++;
                                    }
                                    if (testCount >= totalCount * 0.85)
                                    {
                                        tupple.Add(new Tuple<AttributeType, AttributeType, int, int, string>(attribute, attributeType, testCount, totalCount, uniqueValue));
                                        var exist =
                                            context.RelatedAttributesSimulated.FirstOrDefault(
                                                x =>
                                                    x.AttributeId == attribute.Id &&
                                                    x.RelatedAttributeId == attributeType.Id && x.Value == uniqueValue &&
                                                    x.RelatedValue == uniqueValuesInner);
                                        if (exist == null)
                                        {
                                            context.RelatedAttributesSimulated.Add(new RelatedAttributesSimulated()
                                            {
                                                AttributeId = attribute.Id,
                                                Probability = Math.Round((double)testCount / totalCount, 2),
                                                RelatedAttributeId = attributeType.Id,
                                                RelatedValue = uniqueValuesInner,
                                                Value = uniqueValue
                                            });
                                        }
                                    }
                                }
                            }


                        }
                    }
                    context.SaveChanges();
                    return tupple;
                }
            }
        }

        public class HighchartModel
        {
            public string Name { get; set; }
            public object[] Data { get; set; }
        }
    }
}
