﻿using System;
using System.Collections.Generic;
using System.Linq;
using Domain.DTO;
using System.Globalization;
using Domain.DTO.DialogueModels;

namespace Domain.Implementation
{
    /// <summary>
    /// Static helper class to calculate information gain values
    /// </summary>
    public static class SelectionStrategies
    {
        /// <summary>
        /// 
        /// </summary>
        /// <param name="cases"></param>
        /// <param name="attributes"></param>
        /// <param name="dataSet"></param>
        /// <returns></returns>
        public static AttributeType GetBestAttributeByEntropy(List<DataCase> cases, List<AttributeType> attributes, DataSet dataSet)
        {
            var bestGain = double.MinValue;
            AttributeType bestAttribute = null;
            var dataSetClasses = dataSet.DataClasses;

            foreach (var attribute in attributes)
            {
                var attributeSplit = SplitOnAttribute(attribute, cases);
                var gain = GetGain(cases, attributeSplit, dataSetClasses);

                if (gain > bestGain)
                {
                    bestGain = gain;
                    bestAttribute = attribute;
                }
            }

            return bestAttribute;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="cases"></param>
        /// <param name="childCases"></param>
        /// <param name="dataClasses"></param>
        /// <returns></returns>
        public static double GetGain(List<DataCase> cases, List<List<DataCase>> childCases, List<DataClass> dataClasses)
        {
            var totalEntropy = 0.0;

            foreach (var childCase in childCases)
            {
                totalEntropy += GetEntropy(childCase, dataClasses);
            }

            var casesEntropy = GetEntropy(cases, dataClasses);
            
            return casesEntropy - totalEntropy;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="dataCases"></param>
        /// <param name="dataClasses"></param>
        /// <returns></returns>
        public static double GetEntropy(List<DataCase> dataCases, List<DataClass> dataClasses)
        {
            var entropy = 0.0;
            var totalCases = dataCases.Count;

            foreach (var dataClass in dataClasses)
            {
                var count = dataCases.Count(dataCase => dataCase.DataClass.Id == dataClass.Id);
                var prob = (double) count/totalCases;

                if (count > 0) entropy += -prob*(Math.Log(prob, 2));
            }

            return entropy;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="attribute"></param>
        /// <param name="dataCases"></param>
        /// <returns></returns>
        public static List<List<DataCase>> SplitOnAttribute(AttributeType attribute, List<DataCase> dataCases)
        {
            //Boolean
            if (attribute.Type == 1)
            {
                return SplitOnBoolean(attribute, dataCases);
            }
            
            //Numeric
            if (attribute.Type == 2 || attribute.Type == 3)
            {
                return SplitOnNumeric(attribute, dataCases);
            }

            //Multiple Choice
            if (attribute.Type == 5)
            {
                return SplitOnMulitpleChoice(attribute, dataCases);
            }

            return null;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="attribute"></param>
        /// <param name="dataCases"></param>
        /// <returns></returns>
        public static List<List<DataCase>> SplitOnBoolean(AttributeType attribute, List<DataCase> dataCases)
        {
            var positiveCases = new List<DataCase>();
            var negativeCases = new List<DataCase>();

            foreach (var dataCase in dataCases)
            {
                var attributeValue = dataCase.Attributes.FirstOrDefault(x => x.AttributeType.Id == attribute.Id);
                if (attributeValue != null)
                {
                    if (attributeValue.Value == "true")
                    {
                        positiveCases.Add(dataCase);
                    }
                    else if (attributeValue.Value == "false")
                    {
                        negativeCases.Add(dataCase);
                    }
                }
            }

            return new List<List<DataCase>>{positiveCases, negativeCases};
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="attribute"></param>
        /// <param name="dataCases"></param>
        /// <returns></returns>
        public static List<List<DataCase>> SplitOnNumeric(AttributeType attribute, List<DataCase> dataCases)
        {
            var sortedNumericValues = GetSortedValues(attribute, dataCases);
            if (sortedNumericValues.Count == 0) return new List<List<DataCase>>();

            var splitPoint = GetSplitPoint(sortedNumericValues);

            var split1 = new List<DataCase>();
            var split2 = new List<DataCase>();
            foreach (var dataCase in dataCases)
            {
                var attributeValue = dataCase.Attributes.FirstOrDefault(x => x.AttributeType.Id == attribute.Id);
                if (attributeValue != null && attributeValue.Value != "?")
                {
                    var value = double.Parse(attributeValue.Value.Replace(",", "."), CultureInfo.InvariantCulture);
                    if(value <= splitPoint) 
                        split1.Add(dataCase);
                    else 
                        split2.Add(dataCase);
                }
            }

            return new List<List<DataCase>>{split1, split2};
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="attribute"></param>
        /// <param name="dataCases"></param>
        /// <returns></returns>
        public static List<List<DataCase>> SplitOnMulitpleChoice(AttributeType attribute, List<DataCase> dataCases)
        {
            var splitSets = new List<List<DataCase>>();

            foreach (var option in attribute.MultiSelects)
            {
                var optionSet = new List<DataCase>();
                foreach (var dataCase in dataCases)
                {
                    var attributeValue = dataCase.Attributes.FirstOrDefault(x => x.AttributeType.Id == attribute.Id);
                   
                    if (attributeValue != null && attributeValue.Value == option.Name) optionSet.Add(dataCase);
                }

                if(optionSet.Any()) splitSets.Add(optionSet);
            }

            return splitSets;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="attribute"></param>
        /// <param name="dataCases"></param>
        /// <returns></returns>
        public static List<double> GetSortedValues(AttributeType attribute, List<DataCase> dataCases)
        {
            var numericValues = new List<double>();

            foreach (var dataCase in dataCases)
            {
                var attributeValue = dataCase.Attributes.FirstOrDefault(x => x.AttributeType.Id == attribute.Id);
                if (attributeValue != null && attributeValue.Value != "?")
                {
                    numericValues.Add(double.Parse(attributeValue.Value.Replace(",", "."), CultureInfo.InvariantCulture));
                }
            }

            return numericValues.OrderBy(d => d).ToList();
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="values"></param>
        /// <returns></returns>
        public static double GetSplitPoint(List<double> values)
        {
            var numOfValues = values.Count;

            if (numOfValues == 1)
            {
                return values.ElementAt(0);
            }
            else if (numOfValues%2 == 1)
            {
                var position = (int) Math.Floor(numOfValues / 2.0) - 1;
                return values.ElementAt(position);
            }
            else
            {
                var position1 = (int)Math.Floor(numOfValues / 2.0) - 1;
                var position2 = (int)Math.Floor(numOfValues / 2.0);

                return (values.ElementAt(position2) + values.ElementAt(position1))/2.0;
            }
        }



        public static AttributeType GetBestAttributeByMagic(DataSet dataSet, DataClass currentBestClass, List<ScoreObject> currentBestScores,List<DataCase> retrievalSet, List<DataCase> cases, List<AttributeType> query)
        {
            var maxValue = 0;
            var distinctValues = 0;
            AttributeType returnQuestion = null;

            foreach (var attributeType in dataSet.AttributeTypes)
            {
                var numberOfHits = 0;

                var values = new List<string>();
                if (query.Select(x => x.Id).Contains(attributeType.Id)) continue;

                foreach (var dataCase in retrievalSet)
                {
                    if (dataCase.Attributes.Select(x => x.AttributeType.Id).Contains(attributeType.Id))
                    {
                        if (currentBestClass != null && dataCase.DataClass.Id == currentBestClass.Id)
                        {

                        }
                        else
                        {
                            var attribute = dataCase.Attributes.FirstOrDefault(x => x.AttributeType.Id == attributeType.Id);

                            if (!values.Contains(attribute.Value))
                            {
                                if (currentBestScores == null || !currentBestScores.Any())
                                {
                                    values.Add(attribute.Value);
                                }
                                else if (
                                    currentBestScores.FirstOrDefault()
                                        .Case.Attributes.FirstOrDefault(x => x.AttributeType.Id == attributeType.Id)
                                        .Value != attribute.Value)
                                {
                                    values.Add(attribute.Value);
                                }
                            }

                            numberOfHits++;
                        }
                    }
                }
                if (numberOfHits >= maxValue)
                {
                    if (numberOfHits == maxValue)
                    {
                        var maxDistinctValuesLocal = 0;
                        foreach (var value in values)
                        {
                            var maxDistinctValues =
                                cases.Where(
                                    x =>
                                        x.Attributes.FirstOrDefault(y => y.AttributeType.Id == attributeType.Id)
                                            .Value == value)
                                    .Count();

                            if (maxDistinctValuesLocal < maxDistinctValues)
                                maxDistinctValuesLocal = maxDistinctValues;
                        }
                        if (maxDistinctValuesLocal > distinctValues)
                        {
                            returnQuestion = attributeType;
                            distinctValues = maxDistinctValuesLocal;
                        }
                    }
                    else
                    {
                        if (returnQuestion != null)
                        {
                            if (returnQuestion.Position <= attributeType.Position)
                            {
                                returnQuestion = attributeType;
                            }
                        }
                        else
                        {
                            returnQuestion = attributeType;
                        }
                    }
                }
            }
            return returnQuestion;        
        }

        public static AttributeType GetBestAttributeHeterogenous(DataSet dataSet, List<DataCase> retrievalSet, List<AttributeType> query)
        {
            var bestAttribute = dataSet.AttributeTypes.First();
            var highestFeq = 0;

            var randomAttrs = dataSet.AttributeTypes.OrderBy(x => Guid.NewGuid()).ToList();

            foreach (var attr in randomAttrs)
            {
                var counter = 0;
                if (query.Any(x => x.Id == attr.Id)) continue;
                foreach (var c in retrievalSet)
                {
                    var answer = c.Attributes.FirstOrDefault(x => x.AttributeType.Id == attr.Id);
                    if (answer != null && answer.Value != "?")
                    {
                        counter++;
                    }
                }
                if (counter > highestFeq)
                {
                    highestFeq = counter;
                    bestAttribute = attr;
                }
            }
            if (highestFeq == 0)
            {
                return null;
            }
            return bestAttribute;
        }
    }
}
